/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/control/group',
    ['io.ox/office/tk/utils',
     'io.ox/office/tk/keycodes',
     'io.ox/office/tk/object/triggerobject'
    ], function (Utils, KeyCodes, TriggerObject) {

    'use strict';

    var // DOM event that will cause a 'group:change' event from a group
        INTERNAL_TRIGGER_EVENT = 'private:trigger',

        // all existing groups and their focusable DOM nodes (stored globally  to be able
        // to use a single 'change:focus' listener instead of one listener per group instance)
        focusableNodesMap = {},

        // the unique IDs of the group instances currently focused (as stack, groups may be embedded)
        focusStack = [];

    // global focus handling ==================================================

    function getGroupKey(group) {
        var key = group.getNode().attr('data-key');
        return key ? ('"' + key + '"') : '<none>';
    }

    function registerFocusGroup(group) {
        focusableNodesMap[group.getUid()] = { group: group, nodes: group.getNode() };
    }

    function registerFocusableNodes(group, nodes) {
        var entry = focusableNodesMap[group.getUid()];
        entry.nodes = entry.nodes.add(nodes);
    }

    function unregisterFocusGroup(group) {
        delete focusableNodesMap[group.getUid()];
        focusStack = _(focusStack).without(group.getUid());
    }

    Utils.on('change:focus', function (event, focusNode) {

        // update CSS focus class, trigger event
        function changeFocusState(group, focused) {
            var type = focused ? 'group:focus' : 'group:blur';
            group.getNode().toggleClass(Utils.FOCUSED_CLASS, focused);
            Utils.FocusLogger.log('type="' + type + '" key=' + getGroupKey(group));
            group.trigger(type);
        }

        // returns whether the focus node is inside one of the focusable nodes
        function containsFocusNode(uid) {
            var focusableNodes = focusableNodesMap[uid].nodes;
            return (focusableNodes.filter(focusNode).length > 0) || (focusableNodes.has(focusNode).length > 0);
        }

        // remove all groups from the stack that are not focused anymore
        while ((focusStack.length > 0) && (!focusNode || !containsFocusNode(_.last(focusStack)))) {
            var uid = focusStack.pop();
            // check that group has not been destroyed already
            if (uid in focusableNodesMap) {
                changeFocusState(focusableNodesMap[uid].group, false);
            }
        }

        // push all groups onto the stack that contain the focus node (and are not yet in the stack)
        if (focusNode) {
            _.chain(focusableNodesMap).keys().difference(focusStack).each(function (uid) {
                if (containsFocusNode(uid)) {
                    focusStack.push(uid);
                    changeFocusState(focusableNodesMap[uid].group, true);
                }
            });
        }
    });

    // class Group ============================================================

    /**
     * Creates a container element used to hold a control. All controls shown
     * in view components must be inserted into such group containers. This is
     * the base class for specialized groups and does not add any specific
     * functionality to the inserted controls.
     *
     * Instances of this class trigger the following events:
     * - 'group:change': If the control has been activated in a special way
     *      depending on the type of the control group. The event handler
     *      receives the new value of the activated control.
     * - 'group:cancel': When the focus needs to be returned to the application
     *      (e.g. when the ESCAPE key is pressed, or when a click on a
     *      drop-down button closes the opened drop-down menu).
     * - 'group:show': After the control has been shown or hidden. The event
     *      handler receives the new visibility state.
     * - 'group:enable': After the control has been enabled or disabled. The
     *      event handler receives the new state.
     * - 'group:layout': After child nodes have been inserted into this group
     *      using the method Group.addChildNodes().
     * - 'group:focus': After the group has been focused, by initially focusing
     *      any of its focusable child nodes. As long as the focus remains
     *      inside the group (even if the focus moves to another DOM node in
     *      the group), no further 'group:focus' event will be triggered.
     * - 'group:blur': After the group has lost the browser focus, after
     *      focusing any other DOM node outside the group. As long as the focus
     *      remains inside the group (even if the focus moves to another DOM
     *      node in the group), the 'group:blur' event will not be triggered.
     *
     * @constructor
     *
     * @extends TriggerObject
     *
     * @param {Object} [initOptions]
     *  A map of options to control the properties of the new group. The
     *  following options are supported:
     *  @param {String} [initOptions.classes]
     *      Additional CSS classes that will be set at the root DOM node of
     *      this instance.
     */
    function Group(initOptions) {

        var // self reference
            self = this,

            // create the group container element
            groupNode = $('<div>').addClass('group'),

            // the current value of this control group (must be undefined initially, NOT null)
            groupValue,

            // the current value options of this control group
            groupOptions = null,

            // update handler functions
            updateHandlers = [];

        // base constructor ---------------------------------------------------

        TriggerObject.call(this);

        // private methods ----------------------------------------------------

        /**
         * Keyboard handler for the entire group.
         */
        function keyHandler(event) {
            if (event.keyCode === KeyCodes.ESCAPE) {
                if (event.type === 'keydown') {
                    $.cancelTracking();
                    if (!event.isDefaultPrevented()) { self.triggerCancel(); }
                }
                return false;
            }
        }

        /**
         * Triggers a 'group:cancel' event and suppresses the default action of
         * the passed event, if this group is disabled.
         */
        function disabledHandler(event) {
            if (!self.isEnabled()) {
                event.preventDefault();
                self.triggerCancel();
            }
        }

        // methods ------------------------------------------------------------

        /**
         * Returns the DOM container element for this group as jQuery object.
         */
        this.getNode = function () {
            return groupNode;
        };

        /**
         * Returns the options that have been passed to the constructor.
         *
         * @returns {Object}
         *  The options passed to the constructor; or an empty object.
         */
        this.getOptions = function () {
            return initOptions || {};
        };

        /**
         * Registers the passed update handler function. These handlers will be
         * called from the method Group.setValue() in order of their
         * registration.
         *
         * @param {Function} updateHandler
         *  The update handler function. Will be called in the context of this
         *  group. Receives the following parameters:
         *  (1) {Any} value
         *      The value passed to the method Group.setValue().
         *  (2) {Object|Undefined} options
         *      The options map passed to the method Group.setValue().
         *  (3) {Any} oldValue
         *      The old value of the group.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.registerUpdateHandler = function (updateHandler) {
            updateHandlers.push(updateHandler);
            return this;
        };

        /**
         * Registers an event handler for specific DOM control nodes, that will
         * trigger a 'group:change' event for this group instance. The value of
         * the 'group:change' event will be determined by the passed value
         * resolver function.
         *
         * @param {String} [type]
         *  The type of the event handler that will be bound to the selected
         *  DOM node(s), e.g. 'click' or 'change'. If omitted, the
         *  'group:change' event can only be triggered manually with the method
         *  'Group.triggerChange()'.
         *
         * @param {Object} [options]
         *  A map of options to control the behavior of this method. The
         *  following options are supported:
         *  @param {HTMLElement|jQuery} [options.source]
         *      The DOM node that catches the specified event. May be a single
         *      DOM control element, or any other container element that
         *      catches the events bubbled up from several descending controls.
         *      In this case, the set of source controls can be filtered with a
         *      jQuery selector (option 'selector', see below). If omitted,
         *      defaults to the root node of this group.
         *  @param {String} [options.selector]
         *      If specified, selects the descendant elements of the source
         *      node, which are actually triggering the events. If omitted, the
         *      event handler will be bound directly to the source node.
         *  @param {Function} [options.eventFilter]
         *      A predicate function that can specify to ignore specific DOM
         *      events. Receives the jQuery event object as first parameter,
         *      and must return a Boolean specifying whether to process the
         *      event. If omitted, all DOM events will be processed.
         *  @param {Function} [options.valueResolver]
         *      A callback function that will return the value of this group
         *      instance associated with the source node and/or event type.
         *      Will be called in the context of this group. Receives the
         *      control node that has triggered the event, as jQuery object.
         *      Must return the current value of the control (e.g. the Boolean
         *      state of a toggle button, the value of a list item, or the
         *      current text in a text input field). May return null to
         *      indicate that a 'group:cancel' event should be triggered
         *      instead of a 'group:change' event. If omitted, defaults to the
         *      static method Utils.getControlValue() that extracts a constant
         *      value from the event's source element. The value will be passed
         *      in the 'group:change' event triggered by this instance.
         *  @param {Function} [options.optionsResolver]
         *      A callback function that will return the additional options for
         *      the value of this group instance associated with the source
         *      node and/or event type. Will be called in the context of this
         *      group. Receives the control node that has triggered the event,
         *      as jQuery object. Must return the current value options of the
         *      control. The value options will be passed in the 'group:change'
         *      and 'group:cancel' events triggered by this instance.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.registerChangeHandler = function (type, options) {

            var // the node to attach the event handler to
                sourceNode = Utils.getOption(options, 'source', groupNode),
                // the jQuery selector for ancestor nodes
                selector = Utils.getStringOption(options, 'selector', ''),
                // a predicate function to filter DOM events
                eventFilter = Utils.getFunctionOption(options, 'eventFilter'),
                // the resolver function for the group value to be passed in the 'group:change' event
                valueResolver = Utils.getFunctionOption(options, 'valueResolver', Utils.getControlValue),
                // the resolver function for the additional options to be passed in the group events
                optionsResolver = Utils.getFunctionOption(options, 'optionsResolver', $.noop);

            // handler for all received DOM events
            function eventHandler(event, data) {

                var value = null, valueOptions = null;

                // ignore events if specified by the event filter callback function
                if (_.isFunction(eventFilter) && (eventFilter.call(self, event) !== true)) {
                    return;
                }

                // get value and options from callback functions
                value = self.isEnabled() ? valueResolver.call(self, $(this)) : null;
                valueOptions = self.isEnabled() ? optionsResolver.call(self, $(this)) : undefined;

                // special handling for specific events
                switch (event.type) {
                case 'click':
                    // click events from keyboard: keep focus on button after SPACE or TAB key (see Utils.setButtonKeyHandler() method)
                    if ((event.keyCode === KeyCodes.SPACE) || (event.keyCode === KeyCodes.TAB)) {
                        valueOptions = Utils.extendOptions(valueOptions, { preserveFocus: true });
                    }
                    break;
                case INTERNAL_TRIGGER_EVENT:
                    // internal trigger event may pass own options as separate parameter
                    valueOptions = Utils.extendOptions(valueOptions, data);
                    break;
                }

                if (_.isNull(value)) {
                    self.triggerCancel(valueOptions);
                } else {
                    self.setValue(value, valueOptions);
                    valueOptions = Utils.extendOptions(valueOptions, { source: Utils.getDomNode(sourceNode) });
                    self.trigger('group:change', value, valueOptions);
                }
            }

            // the resulting event type names
            type = _.isString(type) ? (type + ' ' + INTERNAL_TRIGGER_EVENT) : INTERNAL_TRIGGER_EVENT;

            // attach event handler to the node
            if (selector.length > 0) {
                $(sourceNode).on(type, selector, eventHandler);
            } else {
                $(sourceNode).on(type, eventHandler);
            }

            return this;
        };

        /**
         * Registers one or more DOM nodes located outside the root node of
         * this group instance, which contain other control nodes that will be
         * included into the focus handling and generation of 'group:focus' and
         * 'group:blur' events. While the browser focus is located inside the
         * passed nodes, this group will not trigger a 'group:blur' event.
         *
         * @param {HTMLElement|jQuery} nodes
         *  The DOM node(s) to be added to the internal focus handling.
         *
         * @returns {Group}
         *  A reference to this instance.
         */
        this.registerFocusableNodes = function (nodes) {
            registerFocusableNodes(this, nodes);
            return this;
        };

        /**
         * Triggers a 'group:layout' event that causes parent objects to
         * refresh their internal layout according to the new contents and size
         * of this group.
         *
         * @returns {Group}
         *  A reference to this instance.
         */
        this.layout = function () {
            Utils.LayoutLogger.log('type="group:layout" key=' + getGroupKey(this));
            return this.trigger('group:layout');
        };

        /**
         * Inserts the passed HTML mark-up into this group (all existing child
         * nodes will be removed). Triggers a 'group:layout' event notifying
         * all listeners about the changed layout of this group.
         *
         * @param {String} markup
         *  The HTML mark-up to be inserted into this group, as string.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.setChildMarkup = function (markup) {
            groupNode.html(markup);
            return this.layout();
        };

        /**
         * Inserts the passed DOM elements into this group, and triggers a
         * 'group:layout' event notifying all listeners about the changed
         * layout of this group.
         *
         * @param {jQuery} nodes
         *  The nodes to be inserted into this group, as jQuery object.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.addChildNodes = function (nodes) {
            groupNode.append(nodes);
            return this.layout();
        };

        /**
         * Inserts the passed control into this group, and marks it to be
         * included into keyboard focus navigation.
         *
         * @param {jQuery} control
         *  The control to be inserted into this group, as jQuery object.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.addFocusableControl = function (control) {
            return this.addChildNodes(control.addClass(Utils.FOCUSABLE_CLASS));
        };

        /**
         * Returns whether this group contains any focusable controls.
         */
        this.hasFocusableControls = function () {
            return this.getFocusableControls().length > 0;
        };

        /**
         * Returns all control nodes contained in this group that need to be
         * included into keyboard focus navigation.
         */
        this.getFocusableControls = function () {
            return groupNode.find(Utils.FOCUSABLE_SELECTOR + Utils.REALLY_VISIBLE_SELECTOR);
        };

        /**
         * Returns whether this group is currently in focused state, either
         * because the browser focus is located inside the group itself, or it
         * is located in any of the additional DOM nodes that have been
         * registered with the method 'Group.registerFocusableNodes()'.
         */
        this.hasFocus = function () {
            return _.indexOf(focusStack, this.getUid()) >= 0;
        };

        /**
         * Sets the focus to the first control in this group, unless it is
         * already focused.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.grabFocus = function () {
            if (!this.hasFocus()) {
                this.getFocusableControls().first().focus();
            }
            return this;
        };

        /**
         * Returns whether this control group is configured to be visible (via
         * the methods Group.show(), Group.hide(), or Group.toggle()). Does not
         * check the effective visibility depending on the visibility of all
         * parent DOM nodes (see method Group.isReallyVisible() to do that).
         */
        this.isVisible = function () {
            return !groupNode.hasClass(Utils.HIDDEN_CLASS);
        };

        /**
         * Returns whether this control group is effectively visible (it must
         * not be hidden by itself, it must be inside the DOM tree, and all its
         * parent nodes must be visible too).
         */
        this.isReallyVisible = function () {
            return groupNode.is(Utils.REALLY_VISIBLE_SELECTOR);
        };

        /**
         * Displays this control group, if it is currently hidden.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.show = function () {
            return this.toggle(true);
        };

        /**
         * Hides this control group, if it is currently visible.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.hide = function () {
            return this.toggle(false);
        };

        /**
         * Toggles the visibility of this control group.
         *
         * @param {Boolean} [state]
         *  If specified, shows or hides the groups depending on the boolean
         *  value. If omitted, toggles the current visibility of the group.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.toggle = function (state) {

            var // whether to show the group instance
                visible = (state === true) || ((state !== false) && this.isVisible());

            if (this.isVisible() !== visible) {
                groupNode.toggleClass(Utils.HIDDEN_CLASS, !visible);
                Utils.LayoutLogger.log('type="group:show" key=' + getGroupKey(this) + ' state=' + visible);
                this.trigger('group:show', visible);
            }
            return this;
        };

        /**
         * Returns whether this control group is enabled.
         */
        this.isEnabled = function () {
            return groupNode.is(Utils.ENABLED_SELECTOR);
        };

        /**
         * Enables or disables this control group.
         *
         * @param {Boolean} [state=true]
         *  If omitted or set to true, the group will be enabled. Otherwise,
         *  the group will be disabled.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.enable = function (state) {

            var // whether to enable the group instance
                enabled = _.isUndefined(state) || (state === true);

            // enable/disable the entire group node with all its descendants
            if (this.isEnabled() !== enabled) {
                groupNode.toggleClass(Utils.DISABLED_CLASS, !enabled).attr('aria-disabled', !enabled);
                Utils.LayoutLogger.log('type="group:enable" key=' + getGroupKey(this) + ' state=' + enabled);
                this.trigger('group:enable', enabled);
            }
            return this;
        };

        /**
         * Returns the current value of this control group.
         *
         * @returns {Any}
         *  The current value of this control group.
         */
        this.getValue = function () {
            return groupValue;
        };

        /**
         * Updates the controls in this group with the specified value, by
         * calling the registered update handlers.
         *
         * @param {Any} value
         *  The new value to be displayed in the controls of this group.
         *
         * @param {Object} [options]
         *  A map with options that will be passed as second parameter to all
         *  registered update handlers.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.setValue = function (value, options) {
            var oldValue = groupValue;
            if (!_.isUndefined(value)) {
                groupValue = value;
                groupOptions = options;
                _(updateHandlers).each(function (updateHandler) {
                    updateHandler.call(this, value, options, oldValue);
                }, this);
            }
            return this;
        };

        /**
         * Debounced call of the Group.setValue() method with the current value
         * of this group. Can be used to refresh the appearance of the group
         * after changing its content.
         */
        this.refresh = _.debounce(function () {
            // prevent calling method on destroyed object in debounced callback
            if (!self.destroyed) {
                self.setValue(groupValue, groupOptions);
            }
        });

        /**
         * Triggers a special event at the passed DOM control node contained in
         * this Group instance. The group will catch the event, and will
         * trigger a 'group:change' event by itself with the value associated
         * to the DOM control node.
         *
         * @param {HTMLElement|jQuery}
         *  A DOM control node of this group that must have been configured as
         *  source of change events with the Group.registerChangeHandler()
         *  method. If this object is a jQuery collection, uses the first DOM
         *  node it contains.
         *
         * @param {Object} [options]
         *  A map with additional options that will be passed to the
         *  'group:change' event listeners of this class.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.triggerChange = function (control, options) {
            $(control).first().trigger(INTERNAL_TRIGGER_EVENT, options);
            return this;
        };

        /**
         * Triggers a 'group:cancel' event at this Group instance.
         *
         * @param {Object} [options]
         *  A map with additional options that will be passed to the
         *  'group:cancel' event listeners of this class.
         *
         * @returns {Group}
         *  A reference to this group.
         */
        this.triggerCancel = function (options) {
            return this.trigger('group:cancel', options);
        };

        // initialization -----------------------------------------------------

        // register for global focus handling
        registerFocusGroup(this);

        // formatting and tool tip
        groupNode.addClass(Utils.getStringOption(initOptions, 'classes', ''));

        // add event handlers
        self.listenTo(groupNode, 'keydown keypress keyup', keyHandler);
        self.listenTo(groupNode, 'mousedown touchstart dragover drop contextmenu', disabledHandler);

        // destroy all class members on destruction
        this.registerDestructor(function () {
            unregisterFocusGroup(this);
            groupNode.remove();
            self = initOptions = groupNode = updateHandlers = null;
        });

    } // class Group

    // exports ================================================================

    // derive this class from class TriggerObject
    return TriggerObject.extend({ constructor: Group });

});
