/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/popup/listmenu',
    ['io.ox/office/tk/utils',
     'io.ox/office/tk/keycodes',
     'io.ox/office/tk/popup/basemenu'
    ], function (Utils, KeyCodes, BaseMenu) {

    'use strict';

    var // the CSS class for menu section nodes
        SECTION_CLASS = 'section-container';

    // class ListMenu =========================================================

    /**
     * Wraps a pop-up menu element, shown on top of the application window, and
     * relative to an arbitrary DOM node. An item list contains a set of button
     * control elements (the list items) that can be stacked vertically, or can
     * be arranged in a tabular layout. The list items can be highlighted. The
     * items will be grouped into vertically stacked section nodes.
     *
     * Instances of this class trigger the following events (additionally to
     * the events triggered by the base class BaseMenu):
     * - 'create:section'
     *      After a new section node has been created with the method
     *      ListMenu.createSectionNode(). Event handlers receive the following
     *      parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {jQuery} sectionNode
     *          The root node of the new section, as jQuery object.
     *      (3) {String} sectionId
     *          The section identifier.
     *      (4) {Object} [options]
     *          The options passed to the method ListMenu.createSectionNode().
     * - 'create:item'
     *      After a button node for a new list item has been created with the
     *      method ListMenu.createItemNode(). Event handlers receive the
     *      following parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {jQuery} buttonNode
     *          The button element representing the new list item, as jQuery
     *          object.
     *      (3) {String} sectionId
     *          The identifier of the menu section containing the new button.
     *      (4) {Any} value
     *          The value of the new list item.
     *      (5) {Object} [options]
     *          The options passed to the method ListMenu.createItemNode().
     *
     * @constructor
     *
     * @extends BaseMenu
     *
     * @param {Object} [initOptions]
     *  Initial options for the item list menu element. Supports all options
     *  also supported by the base class BaseMenu. Additionally, the following
     *  options are supported:
     *  @param {Function} [initOptions.equality=_.isEqual]
     *      A comparison function that returns whether an arbitrary value
     *      should be considered being equal to the value of a list item. If
     *      omitted, uses _.isEqual() which compares arrays and objects deeply.
     *  @param {String} [initOptions.itemDesign='list']
     *      The design mode of the list item elements:
     *      - 'list': All item elements will be stacked vertically, and will
     *          occupy the complete width of the pop-up menu. Highlighted item
     *          elements will be drawn with a changed background color.
     *      - 'grid': All item elements will be arranged in a tabular layout
     *          per menu section. Highlighted items will be drawn with a thick
     *          border, leaving the background untouched.
     *  @param {Boolean|Function} [initOptions.sortItems=false]
     *      If set to true, the item elements will be sorted inside their menu
     *      section by their label texts, ignoring case; items without text
     *      label will be inserted in no special order. If set to a function,
     *      it must return an ordinal value for each item element. The item
     *      elements will be sorted according to these values, either by number
     *      or lexicographically by strings. The function receives the button
     *      element representing the drop-down menu item as jQuery object, and
     *      will be called in the context of this item list instance. If
     *      omitted or set to false, new item elements will be appended to the
     *      section node.
     *  @param {Number} [options.gridColumns=10]
     *      The default number of columns in each menu section in 'grid' design
     *      mode (see option 'itemDesign' above). This value can be overridden
     *      for single menu sections by passing this option to the method
     *      ListMenu.createSectionNode().
     */
    function ListMenu(initOptions) {

        var // self reference
            self = this,

            // comparator for list item values
            equality = Utils.getFunctionOption(initOptions, 'equality', _.isEqual),

            // the design mode
            design = Utils.getStringOption(initOptions, 'itemDesign', 'list'),

            // functor used to sort the list items
            sortFunctor = Utils.getFunctionOption(initOptions, 'sortItems'),

            // default number of columns in each section node ('grid' design mode)
            defColumns = Utils.getIntegerOption(initOptions, 'gridColumns', 10, 1);

        // base constructor ---------------------------------------------------

        BaseMenu.call(this, Utils.extendOptions({ preferFocusFilter: Utils.SELECTED_SELECTOR }, initOptions));

        // private methods ----------------------------------------------------

        /**
         * Handles 'keydown' events in the root node of the pop-up menu.
         */
        function keyDownHandler(event) {

            var // the list item to be focused
                targetNode = self.getItemNodeForKeyEvent(event, $(event.target));

            // stop event propagation for all supported keys
            if (targetNode.length > 0) {
                targetNode.focus();
                self.scrollToChildNode(targetNode);
                return false;
            }
        }

        // methods ------------------------------------------------------------

        /**
         * Adds a new section to this pop-up menu, if it does not exist yet.
         *
         * @param {String} sectionId
         *  The unique identifier of the menu section.
         *
         * @param {Object} [options]
         *  A map of options to control the appearance of the section. The
         *  following options are supported:
         *  @param {String} [options.label]
         *      If specified, a heading label will be created for the section.
         *  @param {String} [options.classes]
         *      Additional CSS classes that will be added to the section root
         *      node.
         *  @param {Number} [options.gridColumns]
         *      The number of columns in the menu section, if this menu is in
         *      'grid' design mode (see option 'itemDesign' passed to the
         *      constructor). If omitted, the default value for this option as
         *      specified in the constructor options will be used.
         *
         * @returns {jQuery}
         *  The root node of the menu section, as jQuery object.
         */
        this.createSectionNode = function (sectionId, options) {

            var // try to find an existing section node
                sectionNode = this.getSectionNode(sectionId),
                // the header label for the section
                label = null;

            if (sectionNode.length === 0) {

                // create the section root node
                sectionNode = Utils.createContainerNode(SECTION_CLASS + ' group')
                    .attr('data-section', sectionId)
                    .addClass(Utils.getStringOption(options, 'classes', ''))
                    .data('grid-columns', Utils.getIntegerOption(options, 'gridColumns', defColumns, 1));

                // set the label text as attribute (added as :before pseudo element by CSS)
                label = Utils.getStringOption(options, 'label', '');
                if (label.length > 0) {
                    sectionNode.attr('data-section-label', label);
                }

                // insert the new section into the drop-down menu, notify listeners
                this.appendContentNodes(sectionNode);
                this.trigger('create:section', sectionNode, sectionId, options);
            }
            return sectionNode;
        };

        /**
         * Returns the DOM nodes of all menu sections.
         *
         * @returns {jQuery}
         *  The DOM nodes representing all menu sections, as jQuery collection.
         */
        this.getSectionNodes = function () {
            return this.getNode().find('.' + SECTION_CLASS);
        };

        /**
         * Returns the DOM node of the specified menu section.
         *
         * @param {String} sectionId
         *  The unique identifier of the menu section.
         *
         * @returns {jQuery}
         *  The DOM node representing the specified menu section. If the menu
         *  section does not exist, returns an empty jQuery collection.
         */
        this.getSectionNode = function (sectionId) {
            return this.getSectionNodes().filter('[data-section="' + sectionId  + '"]');
        };

        /**
         * Adds a new item element to this list. If the items will be sorted
         * (see the sort options passed to the constructor), the item element
         * will be inserted according to these settings.
         *
         * @param {String} sectionId
         *  The unique identifier of the menu section the new list item will be
         *  inserted into. If a section with this identifier does not exist
         *  yet, it will be created without a header label.
         *
         * @param {Any} value
         *  The unique value associated to the list item. MUST NOT be null or
         *  undefined.
         *
         * @param {Object} [options]
         *  A map of options to control the properties of the button DOM node
         *  representing the new list item. See method Utils.createButton() for
         *  details, except the option 'value' which will be set to the 'value'
         *  parameter passed to this method.
         *
         * @returns {jQuery}
         *  The button element representing the new list item, as jQuery
         *  object.
         */
        this.createItemNode = function (sectionId, value, options) {

            var // the root node of the menu section
                sectionNode = this.createSectionNode(sectionId),
                // all existing item buttons in the current section
                buttonNodes = sectionNode.find(Utils.BUTTON_SELECTOR),
                // create the button element representing the item
                buttonNode = Utils.createButton(Utils.extendOptions(options, { value: value })).addClass(Utils.FOCUSABLE_CLASS),
                // insertion index for sorted items
                sortIndex = -1;

            // inserts the button node in a list
            function insertListItemNode() {
                if ((0 <= sortIndex) && (sortIndex < buttonNodes.length)) {
                    buttonNode.insertBefore(buttonNodes[sortIndex]);
                } else {
                    sectionNode.append(buttonNode);
                }
            }

            // inserts the button node in a table
            function insertGridItemNode() {

                var // the number of columns in the table
                    columns = sectionNode.data('grid-columns'),
                    // the table element containing the grid items
                    tableNode = null,
                    // the last table row
                    rowNode = null;

                // create a new table element for the button if required
                tableNode = sectionNode.find('>table');
                if (tableNode.length === 0) {
                    tableNode = $('<table>').attr('role', 'grid').appendTo(sectionNode);
                }

                // create a new table row element for the button if required
                rowNode = tableNode.find('>tbody>tr').last();
                if ((rowNode.length === 0) || (rowNode.children().last().children().length > 0)) {
                    // IE9 does not allow to write to the 'innerHTML' property for <TR> nodes :-O
                    rowNode = $('<tr role="row">' + Utils.repeatString('<td role="gridcell"></td>', columns) + '</tr>').appendTo(tableNode);
                }

                // insert the new button into the array, and reinsert all buttons into the table
                buttonNodes = buttonNodes.get();
                buttonNodes.splice(sortIndex, 0, buttonNode);
                tableNode.find('>tbody>tr>td').each(function (index) {
                    $(this).append(buttonNodes[index]);
                });
            }

            // find insertion index for sorted items
            if (_.isFunction(sortFunctor)) {
                sortIndex = _.chain(buttonNodes.get())
                    // convert array of button elements to strings returned by sort functor
                    .map(function (node) { return sortFunctor.call(self, $(node)); })
                    // calculate the insertion index of the new list item
                    .sortedIndex(sortFunctor.call(this, buttonNode))
                    // exit the call chain, returns result of sortedIndex()
                    .value();
            } else {
                // else: append to existing items
                sortIndex = buttonNodes.length;
            }

            // insert the new button according to the design mode
            switch (design) {
            case 'list':
                insertListItemNode();
                break;
            case 'grid':
                insertGridItemNode();
            }

            // notify listeners
            this.trigger('create:item', buttonNode, sectionId, value, options);

            return buttonNode;
        };

        /**
         * Returns all button elements representing the menu items of this
         * list.
         *
         * @param {String} [sectionId]
         *  If specified, the result will contain the item elements of that
         *  menu section only.
         *
         * @returns {jQuery}
         *  All button elements representing the menu items in this item list,
         *  or the buttons of a single menu section, as jQuery collection.
         */
        this.getItemNodes = function (sectionId) {
            var node = _.isString(sectionId) ? this.getSectionNode(sectionId) : this.getNode();
            return node.find(Utils.BUTTON_SELECTOR);
        };

        /**
         * Returns the button elements representing the list item with the
         * specified value or selected by the specified predicate function.
         *
         * @param {Any|Function} selector
         *  The value whose list item buttons will be returned, or a predicate
         *  function that receives the button node as jQuery object in the
         *  first parameter, and the list item value in the second parameter,
         *  and that must return a Boolean value whether to include the list
         *  item in the result.
         *
         * @returns {jQuery}
         *  The button elements representing the list items with the specified
         *  value, as jQuery collection.
         */
        this.findItemNodes = function (selector) {
            return (_.isUndefined(selector) || _.isNull(selector)) ? $() : this.getItemNodes().filter(function () {
                var value = Utils.getControlValue($(this));
                return _.isFunction(selector) ? selector.call(self, $(this), value) : equality(selector, value);
            });
        };

        /**
         * Returns the button elements representing all selected list items in
         * this list menu.
         *
         * @returns {jQuery}
         *  The button elements representing all selected list items, as jQuery
         *  collection.
         */
        this.getSelectedItemNodes = function () {
            return Utils.getSelectedButtons(this.getItemNodes());
        };

        /**
         * Selects the specified list item in this menu, and deselects all
         * other list items.
         *
         * @param {Number} index
         *  The zero-based index of the list item to be selected.
         *
         * @param {Object} [options]
         *  A map with additional options controlling the behavior of this
         *  method. The following options are supported:
         *  @param {Boolean} [options.scroll=false]
         *      If set to true, the menu will scroll to the selected item node
         *      automatically.
         *
         * @returns {jQuery}
         *  The button element representing the selected list item, as jQuery
         *  object. If the passed index is invalid, an empty jQuery collection
         *  will be returned.
         */
        this.selectItem = function (index, options) {
            return this.selectItemNode(this.getItemNodes().eq(index), options);
        };

        /**
         * Selects the specified button elements representing one or more list
         * items in this menu, and deselects all other list items.
         *
         * @param {HTMLElement|jQuery} buttonNodes
         *  The list item buttons to be selected, as plain DOM node, or as
         *  jQuery collection that may contain multiple list items.
         *
         * @param {Object} [options]
         *  A map with additional options controlling the behavior of this
         *  method. The following options are supported:
         *  @param {Boolean} [options.scroll=false]
         *      If set to true, the menu will scroll to the selected item node
         *      automatically.
         *
         * @returns {jQuery}
         *  The button elements representing the selected list items, as jQuery
         *  collection.
         */
        this.selectItemNode = function (buttonNode, options) {
            buttonNode = $(buttonNode);
            Utils.selectButtons(this.getItemNodes(), false);
            Utils.selectButtons(buttonNode, true);
            if ((buttonNode.length > 0) && Utils.getBooleanOption(options, 'scroll', false)) {
                this.scrollToChildNode(buttonNode);
            }
            return buttonNode;
        };

        /**
         * Selects the button elements representing the list item with the
         * specified value or selected by the specified predicate function, and
         * deselects all other list items.
         *
         * @param {Any|Function} selector
         *  The value whose list item buttons will be selected, or a predicate
         *  function that receives the button node as jQuery object in the
         *  first parameter, and the list item value in the second parameter,
         *  and that must return a Boolean value whether to select the list
         *  item.
         *
         * @param {Object} [options]
         *  A map with additional options controlling the behavior of this
         *  method. The following options are supported:
         *  @param {Boolean} [options.scroll=false]
         *      If set to true, the menu will scroll to the selected item node
         *      automatically.
         *
         * @returns {jQuery}
         *  The button elements representing the selected list items with the
         *  specified value, as jQuery collection.
         */
        this.findAndSelectItemNodes = function (selector, options) {
            return this.selectItemNode(this.findItemNodes(selector), options);
        };

        /**
         * Returns the target list item to be focused or selected after moving
         * the cursor according to the passed keyboard event.
         *
         * @param {jQuery.Event} keyEvent
         *  The event object caught for a 'keydown' event.
         *
         * @param {HTMLElement|jQuery|Null} buttonNode
         *  The button element representing the source list item the cursor
         *  movement originates from. MUST be a button element representing a
         *  list item of this list menu.
         *
         * @param {Object} [options]
         *  A map with additional options controlling the behavior of this
         *  method. The following options are supported:
         *  @param {Boolean} [options.selectEmpty=false]
         *      If set to true, and the passed button node is invalid (null, or
         *      an empty jQuery collection), selects the first or last list
         *      item instead. All keys in 'down' orientation (CURSOR_DOWN,
         *      PAGE_DOWN) and the HOME key will select the first list item,
         *      all keys in 'up' orientation (CURSOR_UP, PAGE_UP) and the END
         *      key will select the last list item.
         *
         * @returns {jQuery}
         *  The target list item to be focused or selected, as jQuery object.
         *  If the passed key event does not describe any key supported for
         *  cursor movement, or the passed button node is invalid, an empty
         *  jQuery collection will be returned instead.
         */
        this.getItemNodeForKeyEvent = function (keyEvent, buttonNode, options) {

            // returns the number of list items currently visible
            function getListPageSize() {
                var itemHeight = self.getItemNodes().filter(':visible').first().outerHeight();
                return (itemHeight > 0) ? Math.max(1, Math.floor(self.getNode()[0].clientHeight / itemHeight) - 1) : 1;
            }

            // returns the previous or next item node according to the passed distance
            function getNextItemNode(diff, wrap) {

                var // all button nodes in this list
                    buttonNodes = self.getItemNodes(),
                    // index of the focused button
                    index = buttonNodes.index(buttonNode);

                if (index < 0) { return $(); }
                if (wrap) {
                    index += (diff % buttonNodes.length) + buttonNodes.length;
                    index %= buttonNodes.length;
                } else {
                    index = Utils.minMax(index + diff, 0, buttonNodes.length - 1);
                }
                return buttonNodes.eq(index);
            }

            // returns the target item node for 'list' design
            function getItemNodeForList() {

                switch (keyEvent.keyCode) {
                case KeyCodes.UP_ARROW:
                    return getNextItemNode(-1, true);
                case KeyCodes.DOWN_ARROW:
                    return getNextItemNode(1, true);
                case KeyCodes.PAGE_UP:
                    return getNextItemNode(-getListPageSize());
                case KeyCodes.PAGE_DOWN:
                    return getNextItemNode(getListPageSize());
                }

                return $();
            }

            // returns the item node in the previous or next table row in 'grid' design
            function getItemNodeInNextRow(diff) {

                var // the table cell node containing the focused button
                    cellNode = $(Utils.findFarthest(self.getNode(), buttonNode, 'td')),
                    // the table row node containing the focused button
                    rowNode = cellNode.parent(),
                    // all button elements in the focused row
                    buttonNodes = rowNode.children().children(),
                    // all row nodes in all tables
                    rowNodes = self.getSectionNodes().find('>table>tbody>tr'),
                    // the column index of the cell in the row
                    colIndex = cellNode.index(),
                    // the row index of the cell in the table
                    rowIndex = rowNodes.index(rowNode),
                    // the column index of the cell in the table
                    colRatio = (colIndex >= 0) ? (colIndex / buttonNodes.length) : 0;

                if ((rowNodes.length === 0) || (rowIndex < 0)) { return $(); }

                // move focus to the specified row, try to keep column position constant inside row
                rowIndex += (diff % rowNodes.length) + rowNodes.length;
                rowNode = rowNodes.eq(rowIndex % rowNodes.length);
                buttonNodes = rowNode.children().children();
                colIndex = Math.min(Math.floor(colRatio * rowNode.children().length), buttonNodes.length - 1);
                return buttonNodes.eq(colIndex);
            }

            // returns the target item node for 'grid' design
            function getItemNodeForGrid() {

                switch (keyEvent.keyCode) {
                case KeyCodes.LEFT_ARROW:
                    return getNextItemNode(-1, true);
                case KeyCodes.RIGHT_ARROW:
                    return getNextItemNode(1, true);
                case KeyCodes.UP_ARROW:
                    return getItemNodeInNextRow(-1);
                case KeyCodes.DOWN_ARROW:
                    return getItemNodeInNextRow(1);
                }

                return $();
            }

            // convert to jQuery collection
            buttonNode = $(buttonNode);

            // passed button node may be invalid (null, or empty jQuery collection)
            if (buttonNode.length === 0) {

                // fall-back to first/last list item if specified
                if (Utils.getBooleanOption(options, 'selectEmpty', false)) {
                    switch (keyEvent.keyCode) {
                    case KeyCodes.HOME:
                    case KeyCodes.DOWN_ARROW:
                    case KeyCodes.PAGE_DOWN:
                        return this.getItemNodes().first();
                    case KeyCodes.END:
                    case KeyCodes.UP_ARROW:
                    case KeyCodes.PAGE_UP:
                        return this.getItemNodes().last();
                    }
                }

                return $();
            }

            // generic key shortcuts
            switch (keyEvent.keyCode) {
            case KeyCodes.HOME:
                return this.getItemNodes().first();
            case KeyCodes.END:
                return this.getItemNodes().last();
            }

            // special shortcuts dependent on design mode
            switch (design) {
            case 'list':
                return getItemNodeForList();
            case 'grid':
                return getItemNodeForGrid();
            }

            return $();
        };

        /**
         * Scrolls the passed list item node into the visible area of the menu.
         *
         * @param {HTMLElement|jQuery} buttonNode
         *  The list item DOM element that will be made visible by scrolling
         *  the menu root node. If this object is a jQuery collection, uses the
         *  first node it contains.
         *
         * @param {Object} [options]
         *  A map of options to control the scroll action. Supports all options
         *  that are supported by the method Utils.scrollToChildNode(). The
         *  option 'padding' is defaulted to the value 5.
         *
         * @returns {ListMenu}
         *  A reference to this instance.
         */
        this.scrollToChildNode = (function () {
            var baseMethod = _.bind(self.scrollToChildNode, self);
            return function (buttonNode, options) {
                return baseMethod(buttonNode, Utils.extendOptions({ padding: 5 }, options));
            };
        }());

        // initialization -----------------------------------------------------

        // add CSS marker for the item design
        this.getNode().addClass('list-menu').attr('data-design', design);

        // register event handlers
        this.getNode().on('keydown', keyDownHandler);

        // convert ENTER, SPACE, and TAB keys in the menu to click events
        Utils.setButtonKeyHandler(this.getNode(), { tab: true });

        // default sort functor: sort by button label text, case insensitive
        if (Utils.getBooleanOption(initOptions, 'sortItems', false)) {
            sortFunctor = function (buttonNode) {
                var label = Utils.getControlLabel(buttonNode);
                return _.isString(label) ? label.toLowerCase() : '';
            };
        }

        // special handling for tables in 'grid' design
        if (design === 'grid') {
            this.on({
                'popup:beforeshow popup:beforelayout': function () { self.getSectionNodes().find('>table').css('width', 'auto'); },
                'popup:show popup:layout': function () { self.getSectionNodes().find('>table').css('width', ''); }
            });
        }

        this.registerDestructor(function () {
            //if (initOptions) {
            //    delete initOptions.equality;
            //    delete initOptions.sortItems;
            //}
            equality = sortFunctor = initOptions = null;
        });

    } // class ListMenu

    // exports ================================================================

    // derive this class from class BaseMenu
    return BaseMenu.extend({ constructor: ListMenu });

});
